// Copyright (c) 2019, Christoph Peters
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the Karlsruhe Institute of Technology nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#pragma once
#include <vector>
#include <istream>
#include <cstdint>
#include <array>
#include <istream>

/*! Handles a stack of 2D blue noise with a single channel.*/
class BlueNoiseSampler{
public:
	/*! Loads blue noise from the given input stream. The file format is 
		essentially a binary dump of the data in this class.*/
	BlueNoiseSampler(std::istream& rBlueNoiseFile){
		rBlueNoiseFile.read(reinterpret_cast<char*>(&mWidth),sizeof(std::size_t));
		rBlueNoiseFile.read(reinterpret_cast<char*>(&mHeight),sizeof(std::size_t));
		rBlueNoiseFile.read(reinterpret_cast<char*>(&mnTexture),sizeof(std::size_t));
		mPitch=mnTexture*mWidth;
		mvBlueNoise.resize(mWidth*mHeight*mnTexture);
		rBlueNoiseFile.read(reinterpret_cast<char*>(mvBlueNoise.data()),sizeof(std::uint16_t)*mvBlueNoise.size());
		pBlueNoise=mvBlueNoise.data();
		for(std::size_t i=0;i!=8*sizeof(std::size_t);++i){
			if((static_cast<std::size_t>(1)<<(i+1))>mnTexture){
				mTextureIndexMask=(1<<i)-1;
				break;
			}
		}
	}

	/*! Retrieves a blue noise value that is uniform between 0 and 1 for the 
		given random pixel and slice. All given indices are wrapped.*/
	inline float getBlueNoise(std::size_t x,std::size_t y,std::size_t slice) const{
		x=x%mWidth;
		y=y%mHeight;
		slice=slice%mnTexture;
		return BlueNoiseSampler::fixedPointToFloat(pBlueNoise[slice+x*mnTexture+y*mPitch]);
	}

	/*! \return The number of blue noise textures (or slices) in this object.*/
	inline std::size_t getTextureCount() const{
		return mnTexture;
	}
	/*! \return The smallest power of two less than or equal to 
				getTextureCount() minus one. Bitwise and with this value yields 
				a valid slice index.*/
	inline std::size_t getTextureIndexMask() const{
		return mTextureIndexMask;
	}
	/*! Retrieves a pointer to the blue noise that is available for the given 
		pixel. Use fixedPointToFloat() to convert to uniform values from 0 to 
		1. The pointer points to data for slice 0, data for slice i is i 
		entries further. i should be take modulo getTextureCount(). All given 
		indices are wrapped.*/
	inline const uint16_t* getBlueNoiseArray(std::size_t x,std::size_t y) const{
		x=x%mWidth;
		y=y%mHeight;
		return pBlueNoise+(x*mnTexture+y*mPitch);
	}
	/*! Turns a fixed point number as pointed to by the result of 
		getPixelSequence() into a uniform float from zero to one.*/
	static inline float fixedPointToFloat(uint16_t fixedPoint){
		return static_cast<float>(fixedPoint)*(1.0f/65536.0f)+0.5f/65536.0f;
	}

private:
	/*! The width and height of each texture.*/
	std::size_t mWidth,mHeight;
	/*! The number of available blue noise textures.*/
	std::size_t mnTexture;
	/*! Product mnTexture*mWidth.*/
	std::size_t mPitch;
	/*! The blue noise. Dividing by 65536.0f yields uniform values from 0 to 1. 
		See the access functions for the memory layout.*/
	std::vector<std::uint16_t> mvBlueNoise;
	/*! The result of mvBlueNoise.data(), maintained to ensure that slow STL 
		implementations do not distort benchmarks.*/
	const std::uint16_t* pBlueNoise;
	/*! \see getTextureIndexMask() */
	std::size_t mTextureIndexMask;
};
