// Copyright (c) 2019, Christoph Peters
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the Karlsruhe Institute of Technology nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#pragma once
#include "HLSLMath.h"

/*! For our simplified setting, this is all the shading data that we need.*/
struct MinimalistShadingData{
	/*! The world-space position of the shading point.*/
	Float3 position;
	/*! The normalized world-space shading normal.*/
	Float3 normal;
	/*! The normalized world-space direction towards the eye.*/
	Float3 outgoingLightDirection;
	/*! dot(normal,outgoingLightDirection) */
	float lambertOutgoing;
	/*! Linear (i.e. from 0 to 1) roughness parameter for the Disney BRDF.*/
	float linearRoughness;
	/*! \see isPureEmissive */
	Float3 color;
	/*! If this is true, color holds the emitted radiance as RGB. Otherwise, it 
		holds the diffuse albedo as RGB.*/
	bool isPureEmissive;
};


/*! An implementation of the Schlick approximation for the Fresnel term.*/
inline float fresnelSchlick(float fD90,float cosTheta){
	float flipped=1.0f-cosTheta;
	float flippedSquared=flipped*flipped;
	return 1.0f+(fD90-1.0f)*(flippedSquared*flipped*flippedSquared);
}

/*! Implementation of the Disney diffuse BRDF. It agrees with the one in Falcor 
	but can be invoked without having full ShadingData.*/
inline Float3 evaluateDisneyDiffuseBRDF(const MinimalistShadingData& shadingData,Float3 incomingLightDirection){
	// Evaluate various (squared) dot products
	float lambertIncoming=dot(shadingData.normal,incomingLightDirection);
	Float3 halfVector=shadingData.outgoingLightDirection+incomingLightDirection;
	float halfVectorDotIncomingSquared=dot(incomingLightDirection,halfVector);
	halfVectorDotIncomingSquared*=halfVectorDotIncomingSquared;
	halfVectorDotIncomingSquared/=dot(halfVector,halfVector);
	// Evaluate the Fresnel terms and put everything together
	float fD90=mad(halfVectorDotIncomingSquared,2.0f*shadingData.linearRoughness,0.5f);
	float fresnelProduct=fresnelSchlick(fD90,shadingData.lambertOutgoing)*fresnelSchlick(fD90,lambertIncoming);
	return (fresnelProduct/M_PI_FLOAT)*(shadingData.isPureEmissive?Float3(0.0f,0.0f,0.0f):shadingData.color);
}
