// Copyright (c) 2019, Christoph Peters
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the Karlsruhe Institute of Technology nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


/*! \file
	A header that makes the HLSL code for sampling spherical caps compile as C++ 
	code with minimal changes.*/

#pragma once
#include <array>
#include <algorithm>
#include <cmath>

/*! Circle constants in single precision.*/
#define M_PI_FLOAT 3.14159265358979323846f
#define M_PI2_FLOAT 6.28318530717958647692f

/*! A real vector with two entries.*/
struct Float2{
	/* Leaves entries undefined.*/
	inline Float2(){}
	inline Float2(float xIn,float yIn):
	x(xIn),y(yIn){
	}

	float x,y;
};

/*! A real vector with three entries.*/
struct Float3{
	/* Leaves entries undefined.*/
	inline Float3(){}
	inline Float3(float xIn,float yIn,float zIn):
	x(xIn),y(yIn),z(zIn){
	}
	/*! Component-wise multiplication.*/
	inline const Float3& operator*=(float rhs){
		x*=rhs;
		y*=rhs;
		z*=rhs;
		return (*this);
	}
	/*! Component-wise addition.*/
	inline const Float3& operator+=(const Float3& rRHS){
		x+=rRHS.x;
		y+=rRHS.y;
		z+=rRHS.z;
		return (*this);
	}

	float x,y,z;
};

/*! Import some intrinsic functions from cmath into the global namespace.*/
using std::max;
using std::min;
using std::sqrt;
using std::sin;
using std::cos;
using std::abs;
using std::asin;

/*! The dot-product of Euclidean space.*/
inline float dot(const Float3& rLHS,const Float3& rRHS){
	return rLHS.x*rRHS.x+rLHS.y*rRHS.y+rLHS.z*rRHS.z;
}
/*! Reciprocal of the square root.*/
inline float rsqrt(float x){
	return 1.0f/sqrt(x);
}
/*! Scalar multiplication for vectors.*/
inline Float3 operator*(float lhs,const Float3& rRHS){
	return Float3(lhs*rRHS.x,lhs*rRHS.y,lhs*rRHS.z);
}
/*! Component-wise multiplication for vectors.*/
inline Float3 operator*(const Float3& rLHS,const Float3& rRHS){
	return Float3(rLHS.x*rRHS.x,rLHS.y*rRHS.y,rLHS.z*rRHS.z);
}
/*! Element-wise addition for vectors.*/
inline Float3 operator+(const Float3 rLHS,const Float3& rRHS){
	return Float3(rLHS.x+rRHS.x,rLHS.y+rRHS.y,rLHS.z+rRHS.z);
}
/*! Element-wise subtraction for vectors.*/
inline Float3 operator-(const Float3 rLHS,const Float3& rRHS){
	return Float3(rLHS.x-rRHS.x,rLHS.y-rRHS.y,rLHS.z-rRHS.z);
}
/*! A multiply-add operation a*b+c for scalars.*/
inline float mad(float a,float b,float c){
	return a*b+c;
}
/*! An element-wise multiply-add operation a*b+c.*/
inline Float3 mad(float a,const Float3& rB,const Float3& rC){
	return a*rB+rC;
}
/*! Clamps a scalar to the range from zero to one.*/
inline float saturate(float x){
	return (x>1.0f)?1.0f:((x<0.0f)?0.0f:x);
}
/*! Linear interpolation from x to y using s.*/
inline float lerp(float x,float y,float s){
	return (1.0f-s)*x+s*y;
}
/*! Cross-product of two vectors.*/
inline Float3 cross(const Float3& rA,const Float3& rB){
	return Float3(
		rA.y*rB.z-rA.z*rB.y,
		rA.z*rB.x-rA.x*rB.z,
		rA.x*rB.y-rA.y*rB.x
	);
}
/*! Returns a scaled unit-length version of the given vector.*/
inline Float3 normalize(const Float3& rVector){
	return rsqrt(dot(rVector,rVector))*rVector;
}
