// Copyright (c) 2019, Christoph Peters
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the Karlsruhe Institute of Technology nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


// Give PSCMaps.h its M_PI
#define _USE_MATH_DEFINES
#include "SamplingSphericalCaps.h"
#include "PSCMaps.h"
#include "DeferredRenderer.h"
#include "BlueNoiseSampler.h"
#include "DisneyDiffuseBRDF.h"
#include <vector>
#include <iostream>
#include <fstream>
#include <utility>

/*! The floating point type used for Urena's technique.*/
typedef double UrenaScalar;

/*! Represents a spherical light source that emits light isotropically in all 
	directions.*/
struct SphereLight{
	/*! The center of the sphere in world space coordinates.*/
	Float3 mWorldSpaceCenter;
	/*! The radius of the sphere in world space coordinates.*/
	float mRadius;
	/*! Each point on the surface of the sphere emits light isotropically in 
		all directions. This vector gives the light radiance in linear RGB.*/
	Float3 mSurfaceRadiance;
};


/*! The available techniques for sampling spherical caps.*/
enum class CapSamplingTechnique:std::uint8_t{
	/*! Sample proportional to projected solid angle using our technique.*/
	ProjectedSolidAngleOurs,
	/*! Sample proportional to solid angle.*/
	SolidAngle,
	/*! Sample proportional to projected solid angle using the radial technique 
		of Urena et al..*/
	ProjectedSolidAngleUrenaParallel,
	/*! Sample proportional to projected solid angle using the radial technique 
		of Urena et al..*/
	ProjectedSolidAngleUrenaRadial,
	/*! A dummy technique that computes a meaningless direction from the random 
		values. Used to measure the overhead of other operations.*/
	Baseline
};


/*! Gathers all constants that are needed to evaluate the shading function.*/
struct ShadingFunctionConstants{
	/*! The sampling technique used for all light sources.*/
	CapSamplingTechnique mTechnique;
	/*! The number of samples per pixel per light source used to evaluate 
		shading with sphere lights.*/
	std::size_t mnSample;
	/*! The factor applied to radiance values before they are output.*/
	float mExposure;
	/*! The linear roughness parameter for the Disney diffuse BRDF used on all 
		materials in the scene.*/
	float mDisneyDiffuseRoughness;
	/*! Pointer to an array of sphere lights to use for shading.*/
	const SphereLight* mpSphereLight;
	/*! The number of used sphere lights.*/
	std::size_t mnSphereLight;
	/*! The world space position of the camera.*/
	Float3 mWorldSpaceCameraPosition;
};


/*! A helper function to feed Urena's method. Given a surface normal, the 
	vector from the shading point to the center of the sphere light and the 
	radius of the sphere light, it constructs a local frame that meets the 
	requirements of this technique and expresses the spherical cap in terms of 
	angles.*/
inline void getUrenaCoordinateFrame(Float3& rOutTangent,Float3& rOutBitangent,float& rOutAlpha,float& rOutBeta,const Float3& rNormal,const Float3& rRelativeSphereCenter,float radius){
	float invCenterDistance=rsqrt(dot(rRelativeSphereCenter,rRelativeSphereCenter));
	Float3 normalizedCenter=invCenterDistance*rRelativeSphereCenter;
	float normalizedCenterElevation=dot(rNormal,normalizedCenter);
	rOutTangent=mad(-normalizedCenterElevation,rNormal,normalizedCenter);
	float invTangentLength=rsqrt(dot(rOutTangent,rOutTangent));
	rOutTangent*=invTangentLength;
	rOutBitangent=cross(rNormal,rOutTangent);
	rOutAlpha=std::asin(std::min(1.0f,std::max(-1.0f,radius*invCenterDistance)));
	rOutBeta=std::asin(std::min(1.0f,std::max(-1.0f,normalizedCenterElevation)));
}


/*! This functor implements the shading function, i.e. it takes data for a 
	pixel from the G-buffer and turns it into a color. To enable some 
	optimizations, the used sampling technique is passed as compile time 
	constant.*/
template<CapSamplingTechnique technique>
class ShadingFunction{
public:
	ShadingFunction(const ShadingFunctionConstants& rConstantsIn,const BlueNoiseSampler& rBlueNoiseSamplerIn):
	rConstants(rConstantsIn),
	rBlueNoiseSampler(rBlueNoiseSamplerIn){
	}

	/*! Does the shading.*/
	inline Float3 operator()(GBuffer::Pixel pixel,const Float3& rPosition,const Float3& rNormal,const Float3& rColor,bool isEmissive) const;

private:
	const ShadingFunctionConstants& rConstants;
	const BlueNoiseSampler& rBlueNoiseSampler;
};


template<CapSamplingTechnique technique>
inline Float3 ShadingFunction<technique>::operator()(GBuffer::Pixel pixel,const Float3& rPosition,const Float3& rNormal,const Float3& rColor,bool isEmissive) const{
	// For purely emissive pixels, everything is exceptionally simple
	if(isEmissive){
		return rColor;
	}
	// Grab the blue noise
	const uint16_t* pBlueNoise=rBlueNoiseSampler.getBlueNoiseArray(pixel[0],pixel[1]);
	const std::size_t blueNoiseIndexMask=rBlueNoiseSampler.getTextureIndexMask();
	std::size_t iSample=0;
	// Prepare shading with each sphere light
	Float3 finalColor(0.0f,0.0f,0.0f);
	const SphereLight* pSphereLightEnd=rConstants.mpSphereLight+rConstants.mnSphereLight;
	MinimalistShadingData shadingData;
	shadingData.normal=normalize(rNormal);
	shadingData.color=rColor;
	shadingData.position=rPosition;
	shadingData.outgoingLightDirection=normalize(rConstants.mWorldSpaceCameraPosition-shadingData.position);
	shadingData.lambertOutgoing=dot(shadingData.normal,shadingData.outgoingLightDirection);
	shadingData.isPureEmissive=false;
	shadingData.linearRoughness=rConstants.mDisneyDiffuseRoughness;
	// We decide on the technique outside of all loops. This should make it 
	// still easier for the compiler to remove irrelevant code and having 
	// different shading branches for different techniques enables a few micro-
	// optimizations at the expense of minor code duplication.
	if(technique==CapSamplingTechnique::ProjectedSolidAngleOurs){
		for(const SphereLight* pSphereLight=rConstants.mpSphereLight;pSphereLight!=pSphereLightEnd;++pSphereLight){
			ProjectedSphericalCap cap;
			prepareProjectedSphericalCapSampling(cap,shadingData.normal,pSphereLight->mWorldSpaceCenter-shadingData.position,pSphereLight->mRadius);
			if(isProjectedSphericalCapEmpty(cap)){
				iSample+=2*rConstants.mnSample;
				continue;
			}
			for(std::size_t j=0;j!=rConstants.mnSample;++j){
				Float2 blueNoise;
				blueNoise.x=BlueNoiseSampler::fixedPointToFloat(pBlueNoise[(iSample++)&blueNoiseIndexMask]);
				blueNoise.y=BlueNoiseSampler::fixedPointToFloat(pBlueNoise[(iSample++)&blueNoiseIndexMask]);
				float densityOverCosine;
				Float3 sampledDirection=sampleProjectedSphericalCap(densityOverCosine,cap,blueNoise);
				Float3 brdf=evaluateDisneyDiffuseBRDF(shadingData,sampledDirection);
				finalColor+=(1.0f/densityOverCosine)*(brdf*pSphereLight->mSurfaceRadiance);
			}
		}
	}
	else if(technique==CapSamplingTechnique::SolidAngle){
		for(const SphereLight* pSphereLight=rConstants.mpSphereLight;pSphereLight!=pSphereLightEnd;++pSphereLight){
			SphericalCap cap;
			prepareSphericalCapSampling(cap,pSphereLight->mWorldSpaceCenter-shadingData.position,pSphereLight->mRadius);
			for(std::size_t j=0;j!=rConstants.mnSample;++j){
				Float2 blueNoise;
				blueNoise.x=BlueNoiseSampler::fixedPointToFloat(pBlueNoise[(iSample++)&blueNoiseIndexMask]);
				blueNoise.y=BlueNoiseSampler::fixedPointToFloat(pBlueNoise[(iSample++)&blueNoiseIndexMask]);
				Float3 sampledDirection=sampleSphericalCap(cap,blueNoise);
				float cosine=dot(shadingData.normal,sampledDirection);
				if(cosine<=0.0f){
					continue;
				}
				Float3 brdf=evaluateDisneyDiffuseBRDF(shadingData,sampledDirection);
				finalColor+=(cosine*cap.solidAngle)*(brdf*pSphereLight->mSurfaceRadiance);
			}
		}
	}
	else if(technique==CapSamplingTechnique::ProjectedSolidAngleUrenaRadial || technique==CapSamplingTechnique::ProjectedSolidAngleUrenaParallel){
		for(const SphereLight* pSphereLight=rConstants.mpSphereLight;pSphereLight!=pSphereLightEnd;++pSphereLight){
			Float3 tangent,bitangent;
			float alpha,beta;
			getUrenaCoordinateFrame(tangent,bitangent,alpha,beta,shadingData.normal,pSphereLight->mWorldSpaceCenter-shadingData.position,pSphereLight->mRadius);
			PSCM::PSCMaps<UrenaScalar> pscMap;
			pscMap.initialize(alpha,beta,technique==CapSamplingTechnique::ProjectedSolidAngleUrenaRadial);
			if(pscMap.is_invisible()){
				continue;
			}
			for(std::size_t j=0;j!=rConstants.mnSample;++j){
				Float2 blueNoise;
				blueNoise.x=BlueNoiseSampler::fixedPointToFloat(pBlueNoise[(iSample++)&blueNoiseIndexMask]);
				blueNoise.y=BlueNoiseSampler::fixedPointToFloat(pBlueNoise[(iSample++)&blueNoiseIndexMask]);
				UrenaScalar localX,localY;
				pscMap.eval_map(blueNoise.x,blueNoise.y,localX,localY);
				UrenaScalar localZ=std::sqrt(1.0-localX*localX-localY*localY);
				Float3 sampledDirection=static_cast<float>(localX)*tangent+static_cast<float>(localY)*bitangent+static_cast<float>(localZ)*shadingData.normal;
				Float3 brdf=evaluateDisneyDiffuseBRDF(shadingData,sampledDirection);
				finalColor+=static_cast<float>(pscMap.get_area())*(brdf*pSphereLight->mSurfaceRadiance);
			}
		}
	}
	else if(technique==CapSamplingTechnique::Baseline){
		for(const SphereLight* pSphereLight=rConstants.mpSphereLight;pSphereLight!=pSphereLightEnd;++pSphereLight){
			for(std::size_t j=0;j!=rConstants.mnSample;++j){
				Float2 blueNoise;
				blueNoise.x=BlueNoiseSampler::fixedPointToFloat(pBlueNoise[(iSample++)&blueNoiseIndexMask]);
				blueNoise.y=BlueNoiseSampler::fixedPointToFloat(pBlueNoise[(iSample++)&blueNoiseIndexMask]);
				Float3 sampledDirection=normalize(Float3(blueNoise.x,1.0f,blueNoise.y));
				Float3 brdf=evaluateDisneyDiffuseBRDF(shadingData,sampledDirection);
				finalColor+=brdf*pSphereLight->mSurfaceRadiance;
			}
		}
	}
	finalColor*=rConstants.mExposure/rConstants.mnSample;
	return finalColor;
}


/*! Implements the shading function that uses point samples for shading with a 
	diffuse BRDF under illumination through sphere lights. This function, 
	supports all sampling techniques.*/
void shadeWithSphereLights(std::vector<Float3>& rvOutImage,const GBuffer& rGBuffer,const ShadingFunctionConstants& rConstants,const BlueNoiseSampler& rBlueNoiseSampler){
	std::chrono::nanoseconds shadingTime;
	std::size_t nThread=std::thread::hardware_concurrency();
	//nThread=1;
	std::cout<<"Running on "<<nThread<<" threads."<<std::endl;
	if(rConstants.mTechnique==CapSamplingTechnique::ProjectedSolidAngleOurs){
		shadeGBuffer(rvOutImage,rGBuffer,nThread,&shadingTime,ShadingFunction<CapSamplingTechnique::ProjectedSolidAngleOurs>(rConstants,rBlueNoiseSampler));
	}
	else if(rConstants.mTechnique==CapSamplingTechnique::SolidAngle){
		shadeGBuffer(rvOutImage,rGBuffer,nThread,&shadingTime,ShadingFunction<CapSamplingTechnique::SolidAngle>(rConstants,rBlueNoiseSampler));
	}
	else if(rConstants.mTechnique==CapSamplingTechnique::ProjectedSolidAngleUrenaParallel){
		shadeGBuffer(rvOutImage,rGBuffer,nThread,&shadingTime,ShadingFunction<CapSamplingTechnique::ProjectedSolidAngleUrenaParallel>(rConstants,rBlueNoiseSampler));
	}
	else if(rConstants.mTechnique==CapSamplingTechnique::ProjectedSolidAngleUrenaRadial){
		shadeGBuffer(rvOutImage,rGBuffer,nThread,&shadingTime,ShadingFunction<CapSamplingTechnique::ProjectedSolidAngleUrenaRadial>(rConstants,rBlueNoiseSampler));
	}
	else if(rConstants.mTechnique==CapSamplingTechnique::Baseline){
		shadeGBuffer(rvOutImage,rGBuffer,nThread,&shadingTime,ShadingFunction<CapSamplingTechnique::Baseline>(rConstants,rBlueNoiseSampler));
	}
	std::cout<<"Shading took "<<shadingTime.count()*1.0e-9<<" seconds."<<std::endl;
}


/*! Shades a plane with a single spherical light source at different 
	elevations. It either records shaded G-buffers for all techniques or 
	performs a systematic benchmark writing all results to std::cout.*/
void shadePlane(const BlueNoiseSampler& rBlueNoiseSampler,bool runBenchmark){
	// Prepare constants
	ShadingFunctionConstants constants;
	constants.mExposure=1.0f;
	constants.mDisneyDiffuseRoughness=0.5f;
	constants.mnSample=1;
	// Define a benchmark. All combinations of entries in these vectors are 
	// evaluated.
	std::vector<std::pair<float,std::string>> vElevationSuffix={{2.0f,"2.0"},{0.9f,"0.9"},{0.5f,"0.5"},{0.1f,"0.1"},{-0.1f,"-0.1"},{-0.5f,"-0.5"},{-0.9f,"-0.9"}};
	std::vector<std::pair<CapSamplingTechnique,std::string>> vTechnique={{CapSamplingTechnique::Baseline,"Baseline"},{CapSamplingTechnique::SolidAngle,"SolidAngle"},{CapSamplingTechnique::ProjectedSolidAngleOurs,"ProjectedSolidAngleOurs"},{CapSamplingTechnique::ProjectedSolidAngleUrenaParallel,"ProjectedSolidAngleUrenaParallel"},{CapSamplingTechnique::ProjectedSolidAngleUrenaRadial,"ProjectedSolidAngleUrenaRadial"}};
	if(!runBenchmark){
		vTechnique.erase(vTechnique.begin());
	}
	std::vector<std::pair<std::uint32_t,std::uint32_t>> vnLightSample;
	if(runBenchmark){
		vnLightSample={{500,1},{1,500}};
	}
	else{
		vnLightSample={{1,1}};
	}
	// Go through the alternatives
	std::vector<Float3> vImage;
	for(auto elevationSuffix:vElevationSuffix){
		// Set up the sphere light
		SphereLight sphereLight;
		sphereLight.mRadius=1.0f;
		sphereLight.mSurfaceRadiance=Float3(1.0f,1.0f,1.0f);
		sphereLight.mWorldSpaceCenter=Float3(0.0f,elevationSuffix.first,0.0f);
		// Load the G-buffer
		std::string gBufferFilePath="GBuffer/GBufferPlane"+elevationSuffix.second+".dat";
		std::ifstream gBufferFile(gBufferFilePath,std::ios_base::binary);
		if(!gBufferFile.is_open()){
			std::cout<<"Failed to open the G-buffer file: "<<gBufferFilePath<<std::endl;
			return;
		}
		GBuffer gBuffer(gBufferFile);
		constants.mWorldSpaceCameraPosition=gBuffer.getWorldSpaceCameraPosition();
		std::cout<<"Elevation: "<<elevationSuffix.second<<", Diffuse pixel count: "<<gBuffer.getDiffusePixelCount()<<std::endl;
		for(auto technique:vTechnique){
			constants.mTechnique=technique.first;
			for(auto lightSample:vnLightSample){
				std::vector<SphereLight> vSphereLight(lightSample.first,sphereLight);
				constants.mpSphereLight=vSphereLight.data();
				constants.mnSphereLight=vSphereLight.size();
				constants.mnSample=lightSample.second;
				// Do the shading
				std::cout<<"Elevation: "<<elevationSuffix.second<<", Technique: "<<technique.second<<", Light count: "<<lightSample.first<<", Sample count: "<<lightSample.second<<std::endl;
				shadeWithSphereLights(vImage,gBuffer,constants,rBlueNoiseSampler);
				// Dump out the result in a binary format
				if(!runBenchmark){
					std::string outputPath="ShadedPlane/"+technique.second+elevationSuffix.second+".dat";
					std::ofstream imageFile(outputPath,std::ios_base::binary);
					GBuffer::Size resolution=gBuffer.getResolution();
					imageFile.write(reinterpret_cast<char*>(&resolution[0]),sizeof(std::size_t));
					imageFile.write(reinterpret_cast<char*>(&resolution[1]),sizeof(std::size_t));
					imageFile.write(reinterpret_cast<char*>(vImage.data()),sizeof(Float3)*vImage.size());
					std::cout<<"Wrote "<<outputPath<<"."<<std::endl;
				}
			}
		}
	}
}


int main(int nArgument,char* pArgument[]){
	// Load blue noise
	std::ifstream blueNoiseFile("BlueNoiseCube64_64_64.dat",std::ios_base::binary);
	if(!blueNoiseFile.is_open()){
		std::cout<<"Failed to open the blue noise file."<<std::endl;
		return 0;
	}
	BlueNoiseSampler blueNoiseSampler(blueNoiseFile);
	// Either run a benchmark or shade the plane
	if(nArgument==2 && std::string(pArgument[1])=="benchmark"){
		shadePlane(blueNoiseSampler,true);
	}
	else{
		shadePlane(blueNoiseSampler,false);
	}
	std::cout<<"Done."<<std::endl;
	return 0;
}
