// Copyright (c) 2019, Christoph Peters
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the Karlsruhe Institute of Technology nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include "RealTimeRenderer.h"
#include <set>
#include <chrono>
#include <ctime>
#include <fstream>

void RealTimeRenderer::onLoad(SampleCallbacks* pSample,RenderContext* pRenderContext){
	// Create the camera and its controller
	mpCamera=Camera::create();
	mpCamera->setDepthRange(0.03f,1000.0f);
	mpCamera->setPosition(glm::vec3(-1.0f,1.0f,1.0f));
	mFirstPersonCameraController.attachCamera(mpCamera);
	// Load the model and the shader used to visualize the sphere lights
	std::string sphereModelFile="UnitSphere.obj";
	mpSphereModel=Model::createFromFile(sphereModelFile.c_str());
	if(!mpSphereModel){
		msgBox("Failed to load the model at path \""+sphereModelFile+"\".");
		return;
	}
	// Load the shader that is used to shade the spheres
	mpSphereProgram=GraphicsProgram::createFromFile("SphereShader.hlsl","","mainPS");
	mpSphereProgramVariables=GraphicsVars::create(mpSphereProgram->getReflector());
	// Load the shader that is used to render the model to the G-buffer and 
	// bind it
	mpModelProgramWithoutMotionVector=GraphicsProgram::createFromFile("WriteGBuffer.ps.hlsl","","main");
	mpModelProgramVariablesWithoutMotionVector=GraphicsVars::create(mpModelProgramWithoutMotionVector->getReflector());
	mpModelProgramWithMotionVector=GraphicsProgram::createFromFile("WriteGBufferWithMotionVector.ps.hlsl","mainVS","mainPS");
	mpModelProgramVariablesWithMotionVector=GraphicsVars::create(mpModelProgramWithMotionVector->getReflector());
	// Initialize state objects
	mpGraphicsState=GraphicsState::create();
	mpRtState=RtState::create();
	mpRtState->setMaxTraceRecursionDepth(1);
	// Account for the fact that we use reverse z
	DepthStencilState::Desc DepthStencilDescription;
	DepthStencilDescription.setDepthFunc(DepthStencilState::Func::Greater);
	mpGraphicsState->setDepthStencilState(DepthStencilState::create(DepthStencilDescription));
	// Have some blue noise ready
	mpBlueNoiseSampler.reset(new BlueNoiseSampler());
	// Prepare to use temporal antialiasing
	mpTemporalAntialiasing=TemporalAA::create();
	// Load the default scene
	setupScene(mvAvailableScene[miScene].label);
}


void RealTimeRenderer::onResizeSwapChain(SampleCallbacks* pSample,uint32_t width,uint32_t height){
	mpGBufferFboWithMotionVector=Fbo::create();
	mpGBufferFboWithoutMotionVector=Fbo::create();
	mpGBufferDepthStencil=Texture::create2D(width,height,ResourceFormat::D32Float,1,1,nullptr,Resource::BindFlags::DepthStencil|Resource::BindFlags::ShaderResource);
	mpGBufferFboWithMotionVector->attachDepthStencilTarget(mpGBufferDepthStencil);
	mpGBufferFboWithoutMotionVector->attachDepthStencilTarget(mpGBufferDepthStencil);
	uint32_t iTarget=0;
	for(Texture::SharedPtr* ppGBuffer:{&mpGBufferDiffuse,&mpGBufferNormal}){
		(*ppGBuffer)=Texture::create2D(width,height,ResourceFormat::RGB10A2Unorm,1,1,nullptr,Resource::BindFlags::RenderTarget|Resource::BindFlags::ShaderResource);
		mpGBufferFboWithMotionVector->attachColorTarget(*ppGBuffer,iTarget);
		mpGBufferFboWithoutMotionVector->attachColorTarget(*ppGBuffer,iTarget);
		++iTarget;
	}
	mpGBufferMotionVector=Texture::create2D(width,height,ResourceFormat::RG16Float,1,1,nullptr,Resource::BindFlags::RenderTarget|Resource::BindFlags::ShaderResource);
	mpGBufferFboWithMotionVector->attachColorTarget(mpGBufferMotionVector,2);
	mpTargetUAV=Texture::create2D(width,height,ResourceFormat::RGBA16Float,1,1,nullptr,Resource::BindFlags::UnorderedAccess|Resource::BindFlags::ShaderResource);
	for(uint32_t i=0;i!=2;++i){
		mppTAATarget[i]=Texture::create2D(width,height,ResourceFormat::RGBA16Float,1,1,nullptr,Resource::BindFlags::RenderTarget|Resource::BindFlags::ShaderResource);;
		mppTAATargetFbo[i]=Fbo::create();
		mppTAATargetFbo[i]->attachColorTarget(mppTAATarget[i],0);
	}
	// Bind the G-buffer and the render target
	mpGraphicsState->setFbo(nullptr);
	bindGBufferToRayGenerationShader();
	// Update the projection matrix to account for the new resolution. We use 
	// a custom projection matrix for reverse z because sufficient depth precision
	// is crucial with the way our depth buffer is part of the G-buffer.
	float invAspectRatio=static_cast<float>(height)/static_cast<float>(width);
	float fieldOfView=((height*3<width)?0.2f:0.4f)*static_cast<float>(M_PI);
	float transformedFieldOfView=1.0f/std::tan(0.5f*fieldOfView);
	mpCamera->setProjectionMatrix(glm::mat4(
		transformedFieldOfView*invAspectRatio, 0.0f, 0.0f, 0.0f,
		0.0f, transformedFieldOfView, 0.0f, 0.0f,
		0.0f, 0.0f, 0.0f, -1.0f,
		0.0f, 0.0f, mpCamera->getNearPlane(), 0.0f));
}


void RealTimeRenderer::onFrameRender(SampleCallbacks* pSample,RenderContext* pRenderContext,const Fbo::SharedPtr& pTargetFramebuffer){
	// Update the camera
	mFirstPersonCameraController.update();
	mpCamera->beginFrame();
	// Implement animations for and storage of the video
	if(miVideoFrame>=0){
		const std::int32_t nFrame=720;
		if(mvSphereLight.size()>=3){
			float phase=static_cast<float>(M_PI2)*miVideoFrame/static_cast<float>(nFrame)-static_cast<float>(0.5f*M_PI);
			mvSphereLight[2].posW=glm::vec3(std::sqrt(0.5f)*std::cos(phase),std::sin(phase)+0.8f,std::sqrt(0.5f)*std::cos(phase));
		}
		// It takes some time until something we do here ends up in the buffer 
		// that is recorded by a screenshot
		if(miVideoFrame>=3){
			pSample->captureScreen("Video/"+std::string(mSampleProjectedSolidAngle?"PSA/":"SA/")+std::to_string(miVideoFrame-3)+".png",".");
		}
		++miVideoFrame;
		if(miVideoFrame==nFrame+3){
			miVideoFrame=-1;
		}
	}
	// Clear all buffers
	const glm::vec4 clearColor(0.0f,0.0f,0.0f,1.0f);
	pRenderContext->clearFbo(mpGBufferFboWithMotionVector.get(),clearColor,0.0f,0,FboAttachmentType::All);
	pRenderContext->clearFbo(pTargetFramebuffer.get(),clearColor,1.0f,0,FboAttachmentType::All);
	pRenderContext->clearFbo(mppTAATargetFbo[iFrame%2].get(),clearColor,1.0f,0,FboAttachmentType::All);
	pRenderContext->clearUAV(mpTargetUAV->getUAV().get(),clearColor);
	// If there is no model to render, we render nothing at all
	if(!mpModel){
		return;
	}
	// Render the model to the G-buffer
	mpGraphicsState->setFbo(mUseTemporalAntialiasing?mpGBufferFboWithMotionVector:mpGBufferFboWithoutMotionVector);
	mpGraphicsState->setProgram(mUseTemporalAntialiasing?mpModelProgramWithMotionVector:mpModelProgramWithoutMotionVector);
	pRenderContext->setGraphicsState(mpGraphicsState);
	pRenderContext->setGraphicsVars(mUseTemporalAntialiasing?mpModelProgramVariablesWithMotionVector:mpModelProgramVariablesWithoutMotionVector);
	mpModelProgramVariablesWithMotionVector["PerFrameCB"]["gViewportSize"]=glm::vec2(pTargetFramebuffer->getWidth(),pTargetFramebuffer->getHeight());
	ModelRenderer::render(pRenderContext,mpModel,mpCamera.get());
	// Render all of the spheres to the G-buffer
	ConstantBuffer* pSphereConstantBuffer=mpSphereProgramVariables["PerFrameCB"].get();
	for(const SphereLight& rSphereLight:mvSphereLight){
		size_t offset=pSphereConstantBuffer->getVariableOffset("gSphereLight");
		pSphereConstantBuffer->setBlob(&rSphereLight,offset,sizeof(SphereLight));
		pSphereConstantBuffer->setVariable(pSphereConstantBuffer->getVariableOffset("gExposure"),mExposure);
		mpGraphicsState->setProgram(mpSphereProgram);
		pRenderContext->setGraphicsState(mpGraphicsState);
		pRenderContext->setGraphicsVars(mpSphereProgramVariables);
		auto* pSphereInstance=mpSphereModel->getMeshInstance(0,0).get();
		pSphereInstance->setTranslation(rSphereLight.posW,true);
		pSphereInstance->setScaling(glm::vec3(rSphereLight.radius,rSphereLight.radius,rSphereLight.radius));
		ModelRenderer::render(pRenderContext,mpSphereModel,mpCamera.get());
	}
	// Update constants for deferred shading
	ConstantBuffer* pDeferredConstantBuffer=mpDeferredProgramVariables->getGlobalVars()->getConstantBuffer("PerFrameCB").get();
	pDeferredConstantBuffer->setVariable(pDeferredConstantBuffer->getVariableOffset("gExposure"),mExposure);
	pDeferredConstantBuffer->setVariable(pDeferredConstantBuffer->getVariableOffset("gDisneyDiffuseRoughness"),mDisneyDiffuseRoughness);
	for(std::size_t i=0;i!=mvSphereLight.size();++i){
		size_t offset=pDeferredConstantBuffer->getVariableOffset("gpSphereLight["+std::to_string(i)+"]");
		pDeferredConstantBuffer->setBlob(&mvSphereLight[i],offset,sizeof(SphereLight));
	}
	pDeferredConstantBuffer->setVariable(pDeferredConstantBuffer->getVariableOffset("gWorldSpaceCameraPosition"),mpCamera->getPosition());
	pDeferredConstantBuffer->setVariable(pDeferredConstantBuffer->getVariableOffset("gProjectionToWorldSpace"),mpCamera->getInvViewProjMatrix());
	pDeferredConstantBuffer->setVariable(pDeferredConstantBuffer->getVariableOffset("gViewportSize"),glm::uvec2(pTargetFramebuffer->getWidth(),pTargetFramebuffer->getHeight()));
	glm::vec4 viewportTransform;
	viewportTransform.x= 2.0f/pTargetFramebuffer->getWidth();
	viewportTransform.y=-2.0f/pTargetFramebuffer->getHeight();
	viewportTransform.z=0.5f*viewportTransform.x-1.0f;
	viewportTransform.w=0.5f*viewportTransform.y+1.0f;
	pDeferredConstantBuffer->setVariable(pDeferredConstantBuffer->getVariableOffset("gViewportTransform"),viewportTransform);
	mpBlueNoiseSampler->setProgramVariables(mpDeferredProgramVariables->getGlobalVars().get(),!mAnimateBlueNoise);
	// Perform deferred shading with ray tracing
	mpGraphicsState->setFbo(nullptr);
	mpSceneRenderer->renderScene(pRenderContext,mpDeferredProgramVariables,mpRtState,glm::uvec3(pTargetFramebuffer->getWidth(),pTargetFramebuffer->getHeight(),1));
	// Run temporal antialiasing or copy the intermediate UAV into the 
	// swapchain
	if(mUseTemporalAntialiasing){
		mpGraphicsState->setFbo(mppTAATargetFbo[iFrame%2]);
	    mpTemporalAntialiasing->execute(pRenderContext,mpTargetUAV,mppTAATarget[(iFrame+1)%2],mpGBufferMotionVector);
		pRenderContext->blit(mppTAATarget[iFrame%2]->getSRV(),pTargetFramebuffer->getRenderTargetView(0));
	}
	else{
		pRenderContext->blit(mpTargetUAV->getSRV(),pTargetFramebuffer->getRenderTargetView(0));
	}
	// Store the G-buffer if requested
	if(!mGBufferFilePath.empty()){
		std::ofstream gBufferFile(mGBufferFilePath,std::ios_base::binary);
		if(gBufferFile.is_open()){
			if(!storeGBuffer(gBufferFile,pRenderContext)){
				msgBox("Failed to access G-buffer data for writing it to a file.");
			}
		}
		else{
			msgBox("Failed to open the output file for a G-buffer. Please make sure that the output path is valid:\n"+mGBufferFilePath);
		}
		mGBufferFilePath.clear();
	}
	++iFrame;
}


void RealTimeRenderer::onGuiRender(SampleCallbacks* pSample,Gui* pGui){
	// Scene selection
	if(pGui->addDropdown("Scene",mvAvailableScene,miScene)){
		setupScene(mvAvailableScene[miScene].label);
	}
	// Sampling settings
	static const Gui::DropdownList vSamplingTechnique={{0,"Solid angle sampling"},{1,"Our projected solid angle sampling"},{2,"Urena and Georgiev radial"}};
	uint32_t iSamplingTechnique=0;
	if(mSampleProjectedSolidAngle){
		iSamplingTechnique=1;
	}
	if(mSampleProjectedSolidAngleUrena){
		iSamplingTechnique=2;
	}
	if(pGui->addDropdown("Sampling",vSamplingTechnique,iSamplingTechnique)){
		mSampleProjectedSolidAngle=(iSamplingTechnique==1);
		mSampleProjectedSolidAngleUrena=(iSamplingTechnique==2);
		loadRayGenerationShader();
	}
	bool lightOrSampleCountChanged=false;
	int traceShadowRays=mTraceShadowRays?1:0;
	if(pGui->addCheckBox("Trace shadow rays",traceShadowRays)){
		mTraceShadowRays=(traceShadowRays==1);
		lightOrSampleCountChanged=true;
	}
	if(pGui->addIntVar("Sample count",mSampleCount,1,1024)){
		lightOrSampleCountChanged=true;
	}
	pGui->addCheckBox("Animate blue noise",mAnimateBlueNoise);
	// Various rendering settings
	pGui->addFloatVar("Exposure",mExposure,0.0f,200.0f,0.05f,false,"%.2f");
	pGui->addFloatVar("Roughness (Disney diffuse BRDF)",mDisneyDiffuseRoughness,0.0f,1.0f,0.01f,false,"%.2f");
	pGui->addCheckBox("Use temporal antialiasing",mUseTemporalAntialiasing);
	if(mUseTemporalAntialiasing){
		mpTemporalAntialiasing->renderUI(pGui,"Temporal antialiasing");
	}
	// Light controls
	for(uint32_t i=0;i<static_cast<uint32_t>(mvSphereLight.size());){
		if(mvSphereLight[i].renderUI(pGui,i) && mvSphereLight.size()>1){
			mvSphereLight.erase(mvSphereLight.begin()+i);
			lightOrSampleCountChanged=true;
		}
		else{
			++i;
		}
	}
	if(pGui->addButton("Add light")){
		mvSphereLight.push_back(SphereLight(glm::vec3(0.0f),1.0f,glm::vec3(1.0f,1.0f,1.0f)));
		lightOrSampleCountChanged=true;
	}
	if(lightOrSampleCountChanged){
		loadRayGenerationShader();
	}
	// Show a button to store the G-buffer in a simplistic binary format that 
	// can be fed into the offline renderer
	if(pGui->addButton("Store G-buffer")){
		// Construct a file name based on the current time
		std::time_t time=std::chrono::system_clock::to_time_t(std::chrono::system_clock::now());
		char timeString[512];
		std::strftime(timeString,512,"%Y-%m-%d_%H-%M-%S",std::localtime(&time));
		mGBufferFilePath=(std::string("GBuffer")+timeString)+".dat";
	}
	// Show buttons for quicksave and quickload
	if(pGui->addButton("Quick save")){
		quickSave(std::ofstream("QuickSave.dat",std::ios_base::binary));
	}
	if(pGui->addButton("Quick load",true)){
		std::ifstream saveFile("QuickSave.dat",std::ios_base::binary);
		if(saveFile.is_open()){
			quickLoad(saveFile);
		}
	}
}


bool RealTimeRenderer::onKeyEvent(SampleCallbacks* pSample,const KeyboardEvent& keyEvent){
	if(keyEvent.type==KeyboardEvent::Type::KeyPressed){
		if(keyEvent.key==KeyboardEvent::Key::F3){
			quickSave(std::ofstream("QuickSave.dat",std::ios_base::binary));
			return true;
		}
		else if(keyEvent.key==KeyboardEvent::Key::F4){
			std::ifstream saveFile("QuickSave.dat",std::ios_base::binary);
			if(saveFile.is_open()){
				quickLoad(saveFile);
			}
			return true;
		}
		else if(keyEvent.key==KeyboardEvent::Key::F11){
			miVideoFrame=0;
			return true;
		}
	}
	return mFirstPersonCameraController.onKeyEvent(keyEvent);
}


bool RealTimeRenderer::setupScene(const std::string& rSceneName){
	// Attempt to load the quicksave for the scene
	std::string quickSavePath;
	if(!findFileInDataDirectories(rSceneName+"/QuickSave.dat",quickSavePath)){
		msgBox("Failed to locate the quick save for the scene \""+rSceneName+"\".");
		return false;
	}
	std::ifstream quickSaveFile(quickSavePath,std::ios_base::binary);
	if(!quickSaveFile.is_open()){
		msgBox("Failed to open the quick save at path \""+quickSavePath+"\".");
		return false;
	}
	quickLoad(quickSaveFile);
	// Load the model that will be rendered each frame
	std::string modelFile=rSceneName+"/"+rSceneName;
	std::string fullModelPath;
	if(findFileInDataDirectories(modelFile+".fbx",fullModelPath)){
		modelFile+=".fbx";
	}
	else{
		modelFile+=".obj";
	}
	mpModel=RtModel::createFromFile(modelFile.c_str(),RtBuildFlags::FastTrace,Model::LoadFlags::BuffersAsShaderResource);
	if(!mpModel){
		msgBox("Failed to load the model at path \""+modelFile+"\".");
		return false;
	}
	// Prepare the scene and its renderer
	mpScene=RtScene::create(RtBuildFlags::FastTrace);
	mpScene->addModelInstance(ObjectInstance<Model>::create(mpModel,glm::mat4()));
	mpSceneRenderer=RtSceneRenderer::create(mpScene);
	// Create a sampler with anisotropic filtering and use it
	Sampler::Desc samplerDesc;
	samplerDesc.setFilterMode(Sampler::Filter::Linear,Sampler::Filter::Linear,Sampler::Filter::Linear);
	samplerDesc.setMaxAnisotropy(4);
	Sampler::SharedPtr pAnisotropicSampler=Sampler::create(samplerDesc);
	mpModel->bindSamplerToMaterials(pAnisotropicSampler);
	// (Re-)load the ray generation shader that performs deferred shading
	loadRayGenerationShader();
	return true;
}


void RealTimeRenderer::loadRayGenerationShader(){
	// Load the ray generation shader that performs deferred shading
    RtProgram::Desc deferredProgramDescription;
    deferredProgramDescription.addShaderLibrary("SphereLightShader.rt.hlsl").setRayGen("main");
    deferredProgramDescription.addHitGroup(0,"","shadowAnyHit").addMiss(0,"shadowMiss");
	deferredProgramDescription.addDefine("SPHERE_LIGHT_COUNT",std::to_string(mvSphereLight.size()));
	deferredProgramDescription.addDefine("SAMPLE_PROJECTED_SOLID_ANGLE",mSampleProjectedSolidAngle?"1":"0");
	deferredProgramDescription.addDefine("SAMPLE_PROJECTED_SOLID_ANGLE_URENA",mSampleProjectedSolidAngleUrena?"1":"0");
	deferredProgramDescription.addDefine("SAMPLE_COUNT",std::to_string(mSampleCount));
	deferredProgramDescription.addDefine("TRACE_SHADOW_RAYS",std::to_string(mTraceShadowRays));
	mpDeferredProgram=RtProgram::create(deferredProgramDescription,4,8);
	// Prepare constants and bindings for this program
	mpDeferredProgramVariables=RtProgramVars::create(mpDeferredProgram,mpScene);
	mpRtState->setProgram(mpDeferredProgram);
	bindGBufferToRayGenerationShader();
}


void RealTimeRenderer::bindGBufferToRayGenerationShader(){
	mpDeferredProgramVariables->getRayGenVars()->setTexture("gRenderTarget",mpTargetUAV);
	mpDeferredProgramVariables->getRayGenVars()->setTexture("gGBufferDepth",mpGBufferDepthStencil);
	mpDeferredProgramVariables->getRayGenVars()->setTexture("gGBufferDiffuse",mpGBufferDiffuse);
	mpDeferredProgramVariables->getRayGenVars()->setTexture("gGBufferNormal",mpGBufferNormal);
}


bool RealTimeRenderer::storeGBuffer(std::ostream& rBinaryFile,RenderContext* pRenderContext) const{
	std::size_t width=mpGBufferDepthStencil->getWidth();
	std::size_t height=mpGBufferDepthStencil->getHeight();
	rBinaryFile.write(reinterpret_cast<char*>(&width),sizeof(std::size_t));
	rBinaryFile.write(reinterpret_cast<char*>(&height),sizeof(std::size_t));
	glm::mat4 projectionToWorldSpace=mpCamera->getInvViewProjMatrix();
	float ppProjectionToWorldSpace[4][4];
	for(glm::length_t i=0;i!=4;++i){
		for(glm::length_t j=0;j!=4;++j){
			ppProjectionToWorldSpace[i][j]=projectionToWorldSpace[i][j];
		}
	}
	rBinaryFile.write(reinterpret_cast<char*>(&ppProjectionToWorldSpace[0][0]),sizeof(float)*16);
	glm::vec3 cameraPosition=mpCamera->getPosition();
	rBinaryFile.write(reinterpret_cast<char*>(&cameraPosition),sizeof(float)*3);
	for(Texture* pGBuffer:{mpGBufferDepthStencil.get(),mpGBufferDiffuse.get(),mpGBufferNormal.get()}){
		std::vector<uint8> vTextureData=pRenderContext->readTextureSubresource(pGBuffer,0);
		if(vTextureData.size()!=width*height*sizeof(std::uint32_t)){
			return false;
		}
		rBinaryFile.write(reinterpret_cast<char*>(vTextureData.data()),vTextureData.size());
	}
	return true;
}


void RealTimeRenderer::quickSave(std::ostream& rBinaryFile) const{
	// Write the camera transform
	for(glm::vec3 vector:{mpCamera->getPosition(),mpCamera->getUpVector(),mpCamera->getTarget()}){
		rBinaryFile.write(reinterpret_cast<char*>(&vector),sizeof(vector));
	}
	// Write the lights
	std::uint64_t nLight=mvSphereLight.size();
	rBinaryFile.write(reinterpret_cast<char*>(&nLight),sizeof(nLight));
	for(SphereLight light:mvSphereLight){
		rBinaryFile.write(reinterpret_cast<char*>(&light),sizeof(SphereLight));
	}
}


void RealTimeRenderer::quickLoad(std::istream& rBinaryFile){
	// Read the camera transform
	glm::vec3 position,up,target;
	rBinaryFile.read(reinterpret_cast<char*>(&position),sizeof(glm::vec3));
	rBinaryFile.read(reinterpret_cast<char*>(&up),sizeof(glm::vec3));
	rBinaryFile.read(reinterpret_cast<char*>(&target),sizeof(glm::vec3));
	mpCamera->setPosition(position);
	mpCamera->setUpVector(up);
	mpCamera->setTarget(target);
	// Read the lights
	std::size_t nLightOld=mvSphereLight.size();
	std::uint64_t nLight;
	rBinaryFile.read(reinterpret_cast<char*>(&nLight),sizeof(nLight));
	mvSphereLight.resize(nLight,SphereLight(glm::vec3(),0.0f,glm::vec3()));
	for(std::uint64_t i=0;i!=nLight;++i){
		rBinaryFile.read(reinterpret_cast<char*>(&mvSphereLight[i]),sizeof(SphereLight));
	}
	// Update the shader if the light count has changed
	if(mpModel && nLightOld!=mvSphereLight.size()){
		loadRayGenerationShader();
	}
}
