// Copyright (c) 2019, Christoph Peters
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the Karlsruhe Institute of Technology nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#pragma once
#include "Falcor.h"
#include "FalcorExperimental.h"
#include "SphereLight.h"
#include "BlueNoiseSampler.h"
#include <vector>
#include <memory>
#include <ostream>

/*! We use Falcor all over the place and do not want to keep saying that.*/
using namespace Falcor;

/*! This class manages the overall application rendering models with shading 
	based around sampling of spherical caps.*/
class RealTimeRenderer:public Renderer{
public:
	/*! Prepares the entire scene for being rendered.*/
	virtual void onLoad(SampleCallbacks* pSample,RenderContext* pRenderContext) override;
	/*! This override resizes the G-buffer.*/
	virtual void onResizeSwapChain(SampleCallbacks* pSample,uint32_t width,uint32_t height) override;
	/*! Takes care of rendering for the entire scene in one frame.*/
	virtual void onFrameRender(SampleCallbacks* pSample,RenderContext* pRenderContext,const Fbo::SharedPtr& pTargetFramebuffer) override;
	/*! Renders some GUI components.*/
	virtual void onGuiRender(SampleCallbacks* pSample,Gui* pGui) override;
	/*! Forwards keyboard input to the camera controller and additionally 
		supports quick save and quick load.*/
	virtual bool onKeyEvent(SampleCallbacks* pSample,const KeyboardEvent& keyEvent) override;
	/*! Forwards mouse input to the camera controller.*/
	virtual bool onMouseEvent(SampleCallbacks* pSample,const MouseEvent& rMouseEvent) override{
		return mFirstPersonCameraController.onMouseEvent(rMouseEvent);
	}

private:
	/*! This function loads the appropriate model for the scene with the given 
		name and makes it the only model in the scene. It also loads a 
		quicksave with a light and camera setup that mirrors the setups used in 
		the paper. See mvAvailableScene for available scenes.
		\return true if the scene was loaded successfully.*/
	bool setupScene(const std::string& rSceneName);
	/*! (Re)loads the ray-generation shader using the current parameters. Note 
		that the shader uses lots of defines, rather than constants to minimize 
		register pressure for each variant.*/
	void loadRayGenerationShader();
	/*! Binds the current G-buffer to the ray-generation shader.*/
	void bindGBufferToRayGenerationShader();
	/*! This function stores the G-buffer contents into a binary file that can 
		be loaded by the offline renderer.
		\return true iff the buffer was stored successfully.*/
	bool storeGBuffer(std::ostream& rBinaryFile,RenderContext* pRenderContext) const;
	/*! Saves the current state of the camera and of all light sources to the 
		given binary output stream.*/
	void quickSave(std::ostream& rBinaryFile) const;
	/*! Loads the current state of the camera and of all light sources from the 
		given binary input stream.*/
	void quickLoad(std::istream& rBinaryFile);

	/*! The camera used to display the scene.*/
	Camera::SharedPtr mpCamera;
	/*! The sphere light sources that are used for shading.*/
	std::vector<SphereLight> mvSphereLight;
	/*! True to sample spherical caps proportional to projected solid angle, 
		false to sample according to solid angle.*/
	bool mSampleProjectedSolidAngle=true;
	/*! True to use Urena's radial technique for sampling of the projected 
		solid angle.*/
	bool mSampleProjectedSolidAngleUrena=false;
	/*! True if ray tracing should be used to generate shadows, false if 
		visibility should be ignored.*/
	bool mTraceShadowRays=true;
	/*! The number of samples per pixel per light source used to evaluate 
		sphere light shading.*/
	int32_t mSampleCount=1;
	/*! The factor applied to radiance values before they are presented.*/
	float mExposure=10.0f;
	/*! The linear roughness parameter for the Disney diffuse BRDF used on all 
		materials in the scene.*/
	float mDisneyDiffuseRoughness=0.5f;
	/*! Whether temporal antialiasing should be used.*/
	bool mUseTemporalAntialiasing=false;
	/*! Whether the blue noise should be changed each frame.*/
	bool mAnimateBlueNoise=true;
	/*! A list of scenes to feed to imgui. The indices are indices in the 
		vector.*/
	Gui::DropdownList mvAvailableScene={{0,"Head"},{1,"Arcade"},{2,"Plane"}};
	/*! The currently displayed scene (see setupScene()).*/
	std::uint32_t miScene=1;
	/*! The current frame of the video or -1 if no video is being recorded.*/
	std::int32_t miVideoFrame=-1;

	/*! The camera uses common first-person controls.*/
	FirstPersonCameraController mFirstPersonCameraController;
	/*! A triangulated unit sphere used to visualize sphere lights.*/
	Model::SharedPtr mpSphereModel;
	/*! Shaders used to render the sphere lights.*/
	GraphicsProgram::SharedPtr mpSphereProgram;
	/*! Constants and other attributes and bindings used by the shaders for the 
		sphere lights.*/
	GraphicsVars::SharedPtr mpSphereProgramVariables;
	/*! The model that is being displayed.*/
	RtModel::SharedPtr mpModel;
	/*! The scene that contains the above model and maintains the top-level 
		acceleration structure.*/
	RtScene::SharedPtr mpScene;
	/*! A renderer that performs deferred shading with ray tracing based on the 
		scene above.*/
	RtSceneRenderer::SharedPtr mpSceneRenderer;
	/*! Shaders used to render the model.*/
	GraphicsProgram::SharedPtr mpModelProgramWithoutMotionVector,mpModelProgramWithMotionVector;
	/*! Constants and other attributes and bindings used by the shaders for the 
		model.*/
	GraphicsVars::SharedPtr mpModelProgramVariablesWithoutMotionVector,mpModelProgramVariablesWithMotionVector;
	/*! The ray-generation shader that performs deferred shading.*/
	RtProgram::SharedPtr mpDeferredProgram;
	/*! Constants and other attributes and bindings used by the shaders for the 
		deferred shading.*/
	RtProgramVars::SharedPtr mpDeferredProgramVariables;
	/*! State settings used for the device.*/
	GraphicsState::SharedPtr mpGraphicsState;
	/*! The state object used for ray tracing.*/
	RtState::SharedPtr mpRtState;

	/*! Render targets for the G-buffer, specifically for the depth stencil 
		data, the diffuse albedo, the world-space normal and motion vectors.*/
	Texture::SharedPtr mpGBufferDepthStencil,mpGBufferDiffuse,mpGBufferNormal,mpGBufferMotionVector;
	/*! This UAV only exists to be copied into the swap chain.*/
	Texture::SharedPtr mpTargetUAV;
	/*! A framebuffer object for the G-buffer with all textures (optionally 
		except the motion vector) bound.*/
	Fbo::SharedPtr mpGBufferFboWithoutMotionVector,mpGBufferFboWithMotionVector;
	/*! If the next G-buffer that is rendered should be stored to a file, this 
		string holds the file path. Otherwise the string is empty.*/
	std::string mGBufferFilePath;

	/*! An array of blue noise used to get stratified random numbers.*/
	std::unique_ptr<BlueNoiseSampler> mpBlueNoiseSampler;

	/*! A post-processing filter that implements temporal antialiasing.*/
	TemporalAA::SharedPtr mpTemporalAntialiasing;
	/*! Ping pong buffers for the current filtered frame and the previous frame 
		used as source and destination by the TAA filter.*/
	Texture::SharedPtr mppTAATarget[2];
	/*! Framebuffer objects with nothing but mppTAATarget bound.*/
	Fbo::SharedPtr mppTAATargetFbo[2];
	/*! The index of the current frame, used to decide which TAA buffer should 
		be used.*/
	uint32_t iFrame=0;
};
